"""Support processing and manipulation of VCF objects."""

import logging
from contextlib import nullcontext
from pathlib import Path

import pysam
from ga4gh.core import ga4gh_identify
from ga4gh.vrs import models as vrs_models
from ga4gh.vrs import normalize
from ga4gh.vrs.dataproxy import _DataProxy
from ga4gh.vrs.extras.annotator.vcf import FieldName, VcfAnnotator

from anyvar import AnyVar
from anyvar.core.objects import VrsObject

_logger = logging.getLogger(__name__)


class VrsObjectRegistrationBatcher:
    """Handles bulk registration of a large number of VrsObjects in batches"""

    BATCH_SIZE = 10000
    batch_collection: list[VrsObject]
    anyvar_instance: AnyVar

    def __init__(self, anyvar_instance: AnyVar) -> None:
        """Initialize batch collection list and AnyVar instance to use for registration

        :param anyvar_instance: The AnyVar instance to use for registration
        :return: None
        """
        self.batch_collection = []
        self.anyvar_instance: AnyVar = anyvar_instance

    def add_to_batch(self, vrs_object: VrsObject) -> None:
        """Adds a VRS Object to the current batch. If batch size limit is now met, register the whole batch.

        :param vrs_object: The VrsObject to add to the batch
        :return: None
        """
        self.batch_collection.append(vrs_object)
        if len(self.batch_collection) >= self.BATCH_SIZE:
            self.register_batch()

    def register_batch(self) -> None:
        """Registers a batch of vrs objects and resets the collection list to prepare for the next batch"""
        if self.batch_collection:
            self.anyvar_instance.put_objects(self.batch_collection)
            self.batch_collection = []


class VcfRegistrar(VcfAnnotator):
    """Custom implementation of annotator class from VRS-Python. Rewrite some methods
    and values in order to enable use of existing AnyVar translator.
    """

    vrs_object_registration_batcher: VrsObjectRegistrationBatcher

    def __init__(self, data_proxy: _DataProxy, **kwargs) -> None:  # noqa: D107
        av: AnyVar | None = kwargs.get("av")
        if av is None:
            raise ValueError
        self.av: AnyVar = av
        self.vrs_object_registration_batcher = VrsObjectRegistrationBatcher(self.av)
        super().__init__(data_proxy)

    def on_vrs_object(
        self,
        vcf_coords: str,
        vrs_allele: vrs_models.Allele,
        **kwargs,
    ) -> vrs_models.Allele | None:
        """Adds the VRS object to the batcher, which handles bulk registration"""
        self.vrs_object_registration_batcher.add_to_batch(vrs_allele)
        return vrs_allele

    def on_vrs_object_collection(  # noqa: D102
        self,
        vrs_alleles_collection: list[vrs_models.Allele] | None,
        **kwargs,
    ) -> None:
        pass

    def raise_for_output_args(self, output_vcf_path: Path | None, **kwargs) -> None:  # noqa: D102
        pass

    def annotate(
        self,
        input_vcf_path: Path,
        output_vcf_path: Path | None = None,
        vrs_attributes: bool = False,
        assembly: str = "GRCh38",
        compute_for_ref: bool = True,
        require_validation: bool = True,
        **kwargs,
    ) -> None:
        """Calls the parent 'annotate' function, then adds the final batch of VRS objects to the database"""
        super().annotate(
            input_vcf_path,
            output_vcf_path,
            vrs_attributes,
            assembly,
            compute_for_ref,
            require_validation,
            **kwargs,
        )

        # register the final batch of vrs objects (since the last batch will likely be smaller than the batch size limit)
        self.vrs_object_registration_batcher.register_batch()


class RequiredAnnotationsError(Exception):
    """Raise when encountering incomplete or invalid VRS annotations"""


def _raise_for_missing_vcf_annotations(vcf: pysam.VariantFile) -> None:
    """Check whether all required VRS annotations are present on a provided VCF

    :param vcf: file to check
    :return: None if successful
    :raise: RequiredAnnotationsError if provided VCF lacks required annotations
    """
    field_names = {name.value for name in FieldName}
    if not all(n in vcf.header.info for n in field_names):
        raise RequiredAnnotationsError(
            "VCF missing some required INFO fields: %s",
            field_names - set(vcf.header.info.keys()),
        )


def register_existing_annotations(
    av: AnyVar, file_path: Path, assembly: str, require_validation: bool = False
) -> Path | None:
    """Extract VRS allele parameters from a previously-annotated VCF and register them.

    :param av: AnyVar instance
    :param file_path: path to VCF
    :param assembly: assembly used during VCF writing and annotation. Probably one of
        ``{"GRCh37", "GRCh38"}``
    :param require_validation: whether to check that annotated ID is correct
    :return:  Path to ID conflict file, if requested, or None otherwise
    :raise: ValueError if input VCF lacks required annotations
    """
    vrs_object_registration_batcher = VrsObjectRegistrationBatcher(av)

    _logger.info("Registering existing annotations from VCF at %s", file_path)
    variantfile = pysam.VariantFile(filename=str(file_path), mode="r")
    _raise_for_missing_vcf_annotations(variantfile)

    if require_validation:
        conflict_logfile_path = file_path.parent / f"{file_path.name}_conflictlog"
        cm = conflict_logfile_path.open("w")
    else:
        # conflicts will be ignored, so no need to keep track of them
        conflict_logfile_path = None
        cm = nullcontext(None)

    with cm as conflict_logfile:
        if conflict_logfile:
            conflict_logfile.write(
                "vrs_id,assembly,chrom,pos,start,end,state,new_vrs_id\n"
            )
        for record in variantfile:
            if not all(
                [
                    FieldName.IDS_FIELD in record.info,
                    FieldName.STARTS_FIELD in record.info,
                    FieldName.ENDS_FIELD in record.info,
                    FieldName.STATES_FIELD in record.info,
                    FieldName.LENGTHS_FIELD in record.info,
                    FieldName.REPEAT_SUBUNIT_LENGTHS_FIELD in record.info,
                ]
            ):
                continue
            sequence = f"{assembly}:{record.chrom}"
            refget_accession = av.translator.dp.derive_refget_accession(sequence)
            if not refget_accession:
                _logger.warning(
                    "Unable to acquire refget accession for constructed sequence identifier %s at pos %s",
                    sequence,
                    record.pos,
                )
                continue
            for vrs_id, start, end, state, length, repeat_subunit_length in zip(
                record.info[FieldName.IDS_FIELD],
                record.info[FieldName.STARTS_FIELD],
                record.info[FieldName.ENDS_FIELD],
                record.info[FieldName.STATES_FIELD],
                record.info[FieldName.LENGTHS_FIELD],
                record.info[FieldName.REPEAT_SUBUNIT_LENGTHS_FIELD],
                strict=True,
            ):
                if vrs_id == ".":
                    continue
                true_state = "" if state == "." else state
                if isinstance(true_state, str):
                    true_state = vrs_models.sequenceString(true_state)

                seq_ref = vrs_models.SequenceReference(refgetAccession=refget_accession)
                location = vrs_models.SequenceLocation(
                    sequenceReference=seq_ref, start=start, end=end
                )
                location_id = ga4gh_identify(location)
                location.id = location_id

                if length != "." and repeat_subunit_length != ".":
                    sequence_expression = vrs_models.ReferenceLengthExpression(
                        length=length,
                        repeatSubunitLength=repeat_subunit_length,
                        sequence=true_state,
                    )
                else:
                    sequence_expression = vrs_models.LiteralSequenceExpression(
                        sequence=true_state
                    )

                allele = vrs_models.Allele(location=location, state=sequence_expression)
                allele = normalize(allele, av.translator.dp)
                new_vrs_id = ga4gh_identify(allele)
                if conflict_logfile and new_vrs_id != vrs_id:
                    _logger.debug(
                        "Annotated ID %s conflicts with newly-calculated ID %s for variation %s:%s %s-%s (state: %s)",
                        vrs_id,
                        new_vrs_id,
                        assembly,
                        record.chrom,
                        start,
                        end,
                        sequence_expression,
                    )
                    conflict_logfile.write(
                        f"{vrs_id},{assembly},{record.chrom},{record.pos},{start},{end},{sequence_expression},{new_vrs_id}\n"
                    )
                vrs_object_registration_batcher.add_to_batch(allele)

            # register the final batch of alleles (since the last batch will likely be smaller than the batch size limit)
            vrs_object_registration_batcher.register_batch()

    return conflict_logfile_path



"""Provide API routes relating to search operations"""

from http import HTTPStatus
from typing import Annotated

from fastapi import APIRouter, HTTPException, Query, Request

from anyvar.anyvar import AnyVar
from anyvar.restapi.schema import SearchResponse

search_router = APIRouter()


@search_router.get(
    "/search",
    response_model_exclude_none=True,
    operation_id="searchVariations",
    summary="Search for registered variations by genomic region",
    description="Return all variants with start and end positions that fall within the provided start and end arguments (inclusive).",
)
def search_variations(
    request: Request,
    accession: Annotated[
        str,
        Query(
            ...,
            description='Sequence accession identifier (for example: `"ga4gh:SQ.8_liLu1aycC0tPQPFmUaGXJLDs5SbPZ5"`)',
            examples=["ga4gh:SQ.8_liLu1aycC0tPQPFmUaGXJLDs5SbPZ5"],
        ),
    ],
    start: Annotated[
        int,
        Query(..., description="Start position for genomic region", examples=[2781631]),
    ],
    end: Annotated[
        int,
        Query(..., description="End position for genomic region", examples=[2781758]),
    ],
) -> SearchResponse:
    """Perform genomic coordinate-based search over all registered variations."""
    av: AnyVar = request.app.state.anyvar
    try:
        if accession.startswith("ga4gh:"):
            ga4gh_id = accession
        else:
            ga4gh_id = av.translator.get_sequence_id(accession)
    except KeyError as e:
        raise HTTPException(
            status_code=HTTPStatus.NOT_FOUND,
            detail="Unable to dereference provided accession ID",
        ) from e

    alleles = []
    if ga4gh_id:
        try:
            refget_accession = ga4gh_id.split("ga4gh:")[-1]
            alleles = av.object_store.search_alleles(refget_accession, start, end)
        except NotImplementedError as e:
            raise HTTPException(
                status_code=HTTPStatus.NOT_IMPLEMENTED,
                detail="Search not implemented for current storage backend",
            ) from e

    return SearchResponse(variations=alleles)



"""Provide router for operations on stored objects"""

import json
import logging
from http import HTTPStatus
from typing import Annotated, cast

from fastapi import APIRouter, Body, HTTPException, Request, Response
from fastapi.params import Path
from fastapi.responses import JSONResponse, StreamingResponse
from ga4gh.vrs.dataproxy import DataProxyValidationError
from hgvs.exceptions import HGVSParseError
from pydantic import StrictStr

from anyvar.anyvar import AnyVar, ObjectNotFoundError
from anyvar.core import metadata, objects
from anyvar.mapping import liftover
from anyvar.restapi.schema import (
    AddAnnotationRequest,
    AddAnnotationResponse,
    AddMappingRequest,
    AddMappingResponse,
    GetAnnotationResponse,
    GetMappingResponse,
    GetObjectResponse,
    RegisterVariationResponse,
    TranslationResult,
    VariationRequest,
)
from anyvar.storage.base import IncompleteVrsObjectError
from anyvar.translate.base import TranslationError, Translator

_logger = logging.getLogger(__name__)

objects_router = APIRouter()


def _get_vrs_object(
    av: AnyVar, vrs_object_id: str, object_type: type[objects.VrsObject] | None = None
) -> objects.VrsObject:
    """Get VRS variation given VRS ID

    :param av: AnyVar instance
    :param vrs_object_id: VRS Object ID to retrieve
    :param object_type: (Optional) The type of object to retrieve
    :raises HTTPException: If no VRS object ID found
    :return: VrsObject
    """
    try:
        return av.get_object(vrs_object_id, object_type)
    except KeyError as e:
        raise HTTPException(
            status_code=HTTPStatus.NOT_FOUND,
            detail=f"VRS Object {vrs_object_id} not found",
        ) from e


async def parse_and_rebuild_response(
    response: StreamingResponse,
) -> tuple[dict, Response]:
    """Convert a `Response` object to a dict, then re-build a new Response object (since parsing exhausts the Response `body_iterator`).

    :param response: the `Response` object to parse
    :return: a tuple with a dictionary representation of the Response and a new `Response` object
    """
    response_chunks: list[bytes] = [
        cast(bytes, chunk) async for chunk in response.body_iterator
    ]
    response_body_encoded = b"".join(response_chunks)
    response_body = response_body_encoded.decode("utf-8")
    response_json = json.loads(response_body)

    new_response = JSONResponse(
        content=response_json,
        status_code=response.status_code,
        media_type=response.media_type,
    )

    return (response_json, new_response)


VARIATION_EXAMPLE_PAYLOAD = {
    "definition": "NC_000007.13:g.36561662_36561663del",
    "input_type": "Allele",
    "copies": 0,
    "copy_change": "complete genomic loss",
    "assembly_name": None,
}


_variation_request_body = Body(
    description='Variation description, including (at minimum) a `definition` property. Can provide optional `input_type` if the expected output representation type is known, as well as an assembly_name (e.g.,"GRCh37" or "GRCh38"). If representing copy number, provide `copies` or `copy_change`.',
    examples=[VARIATION_EXAMPLE_PAYLOAD],
)


def _translate_variation(
    tlr: Translator, variation_request: VariationRequest
) -> TranslationResult:
    """Perform variant translation

    :param tlr: Translator instance
    :param variation_request: Input variation request to translate
    :return: TranslationResult object with translated variation, if translation is
        successful. Otherwise, return error message
    """
    definition = variation_request.definition

    try:
        translated_variation = tlr.translate_variation(
            definition, **variation_request.model_dump(mode="json")
        )
        return TranslationResult(variation=translated_variation)
    except DataProxyValidationError as e:
        return TranslationResult(error=str(e))
    except HGVSParseError:
        return TranslationResult(
            error=f'Unable to parse HGVS expression "{definition}"'
        )
    except NotImplementedError:
        return TranslationResult(
            error=f"Variation class for {definition} is currently unsupported."
        )
    except TranslationError:
        return TranslationResult(error=f'Unable to translate "{definition}"')


def _handle_translation_request(
    tlr: Translator, var_req: VariationRequest
) -> objects.VrsVariation:
    """Perform variant translation and convert known exceptions to appropriate HTTP responses

    :param tlr: Translator instance
    :param var_req: request object relayed to variation endpoint
    :return: VRS variation instance
    :raise HTTPException: return 422 response if
       * Variant definition cannot be translated
       * Reference base in gnomad/VCF-style expression fails to validate
       * translator returns not-implemented variation type
    """
    translation_result = _translate_variation(tlr, var_req)
    if translation_result.error:
        raise HTTPException(
            status_code=HTTPStatus.UNPROCESSABLE_ENTITY,
            detail=translation_result.error,
        )

    return translation_result.variation  # type: ignore


def _register_variations(
    av: AnyVar, variation_requests: list[VariationRequest]
) -> list[RegisterVariationResponse]:
    """Bulk register variations

    :param av: AnyVar instance
    :param variation_requests: Input variation requests to register
    :return: List of RegisterVariationResponse objects in the same order as the input.
        Variations that fail translation are not registered and are returned with null
        `object` and `object_id` fields. Registration or liftover failure messages may
        also be included in the `messages` field.
    """
    translation_results: list[TranslationResult] = []
    variations_to_store: list[objects.VrsObject] = []

    for variation_request in variation_requests:
        translation_result = _translate_variation(av.translator, variation_request)
        translation_results.append(translation_result)

        if translation_result.variation:
            variations_to_store.append(translation_result.variation)

    if variations_to_store:
        av.put_objects(variations_to_store)

    responses: list[RegisterVariationResponse] = []

    for variation_request, translation_result in zip(
        variation_requests, translation_results, strict=True
    ):
        if not translation_result.variation:
            responses.append(
                RegisterVariationResponse(
                    input_variation=variation_request,
                    messages=[translation_result.error]
                    if translation_result.error
                    else [],
                )
            )
            continue

        # add variant metadata
        av.create_timestamp_annotation_if_missing(translation_result.variation.id)  # type: ignore (ID guaranteed to be present)
        messages: list[str] = (
            liftover.add_liftover_mapping(
                variation=translation_result.variation,
                storage=av.object_store,
                dataproxy=av.translator.dp,
            )
            or []
        )

        responses.append(
            RegisterVariationResponse(
                input_variation=variation_request,
                object=translation_result.variation,
                object_id=translation_result.variation.id
                if translation_result.variation
                else None,
                messages=messages,
            )
        )

    return responses


@objects_router.put(
    "/variation",
    response_model_exclude_none=True,
    summary="Register a new allele or copy number object",
    description="Provide a variation definition  to be normalized and registered with AnyVar. A complete VRS Allele or Copy Number object and ID is returned.",
)
def register_variation(
    request: Request,
    variation: Annotated[VariationRequest, _variation_request_body],
) -> RegisterVariationResponse:
    """Register a variation based on a provided description or reference."""
    av: AnyVar = request.app.state.anyvar

    responses: list[RegisterVariationResponse] = _register_variations(av, [variation])
    return responses[0]


@objects_router.put(
    "/variations",
    response_model_exclude_none=True,
    summary="Bulk register alleles or copy number objects",
    description="Provide a list of variation definitions to be normalized and registered with AnyVar. The response contains one result per input, in the same order. Variations that fail translation are not registered and are returned with null `object` and `object_id` fields. Registration or liftover failure messages may also be included in the `messages` field.",
)
def register_variations(
    request: Request,
    variations: Annotated[
        list[VariationRequest], Body(description="List of variations to register")
    ],
) -> list[RegisterVariationResponse]:
    """Register multiple variations based on provided descriptions or references."""
    av: AnyVar = request.app.state.anyvar
    return _register_variations(av, variations)


PUT_VRS_VARIATION_EXAMPLE_PAYLOAD = {
    "location": {
        "end": 87894077,
        "start": 87894076,
        "sequenceReference": {
            "refgetAccession": "SQ.ss8r_wB0-b9r44TQTMmVTI92884QvBiB",
            "type": "SequenceReference",
        },
        "type": "SequenceLocation",
    },
    "state": {"sequence": "T", "type": "LiteralSequenceExpression"},
    "type": "Allele",
}


@objects_router.put(
    "/vrs_variation",
    summary="Register a VRS variation",
    description="Provide a valid VRS variation object to be registered with AnyVar. Returns a fully-identified VRS object.",
    response_model_exclude_none=True,
)
def register_vrs_variation(
    request: Request,
    variation: Annotated[
        objects.VrsVariation,
        Body(
            description="Valid VRS object.",
            examples=[PUT_VRS_VARIATION_EXAMPLE_PAYLOAD],
        ),
    ],
) -> RegisterVariationResponse:
    """Register a complete VRS variation object.

    No additional formatting or normalization is performed. IDs are added if not provided.
    """
    av: AnyVar = request.app.state.anyvar
    input_variation = variation
    try:
        av.put_objects([variation])
    except IncompleteVrsObjectError:
        variation = objects.recursive_identify(variation)
        av.put_objects([variation])

    liftover_messages = liftover.add_liftover_mapping(
        variation, av.object_store, av.translator.dp
    )

    return RegisterVariationResponse(
        input_variation=input_variation,
        object=variation,
        object_id=variation.id,
        messages=liftover_messages or [],
    )


@objects_router.post(
    "/variation",
    response_model_exclude_none=True,
    summary="Retrieve a registered VRS allele or copy number variation",
    description="Provide a variation definition to be normalized and searched for in AnyVar",
)
def get_variation(
    request: Request,
    variation: Annotated[VariationRequest, _variation_request_body],
) -> GetObjectResponse:
    """Search for registered variation"""
    av: AnyVar = request.app.state.anyvar
    translated_variation = _handle_translation_request(av.translator, variation)
    vrs_id: str = translated_variation.id  # type: ignore
    _ = _get_vrs_object(av, vrs_id)  # raise NOT_FOUND for vrs_id not present in DB
    return GetObjectResponse(messages=[], data=translated_variation)


@objects_router.get(
    "/object/{vrs_id}",
    response_model_exclude_none=True,
    operation_id="getVariation",
    summary="Retrieve a VRS object",
    description="Gets a VRS object by ID. May return any supported type of VRS Object.",
)
def get_object_by_id(
    request: Request,
    vrs_id: Annotated[StrictStr, Path(..., description="VRS ID for object")],
) -> GetObjectResponse:
    """Get registered VRS object given its VRS ID."""
    av: AnyVar = request.app.state.anyvar
    vrs_object: objects.VrsObject = _get_vrs_object(av, vrs_id)
    return GetObjectResponse(messages=[], data=vrs_object)


@objects_router.post(
    "/object/{vrs_id}/annotations",
    response_model_exclude_none=True,
    summary="Add annotation to a VRS Object",
    description="Provide an annotation to associate with a VRS object. The object MUST be registered with AnyVar before it can be annotated.",
)
def add_object_annotation(
    request: Request,
    vrs_id: Annotated[
        StrictStr, Path(..., description="VRS ID of variation to annotate")
    ],
    annotation_request: Annotated[
        AddAnnotationRequest,
        Body(
            description="Annotation to associate with the variation",
        ),
    ],
) -> AddAnnotationResponse:
    """Store an annotation for a VRS Object."""
    # Look up the VRS Object from the AnyVar store
    av: AnyVar = request.app.state.anyvar
    vrs_object: objects.VrsObject = _get_vrs_object(av, vrs_id)

    # Add the annotation to the database
    annotation_id: int | None = None
    try:
        annotation = metadata.Annotation(
            object_id=vrs_object.id,  # pyright: ignore[reportArgumentType] - VRS Objects from the DB will never NOT have an ID
            annotation_type=annotation_request.annotation_type,
            annotation_value=annotation_request.annotation_value,
        )
        annotation_id = av.put_annotation(annotation)
    except ValueError as e:
        _logger.exception(
            "Failed to add annotation `%s` on VRS Object `%s`",
            annotation_request,
            vrs_id,
        )
        raise HTTPException(
            status_code=HTTPStatus.INTERNAL_SERVER_ERROR,
            detail=f"Failed to add annotation: {annotation_request}",
        ) from e

    return AddAnnotationResponse(
        object=vrs_object,
        object_id=vrs_id,
        annotation_type=annotation_request.annotation_type,
        annotation_value=annotation_request.annotation_value,
        annotation_id=annotation_id,
    )


@objects_router.get(
    "/object/{vrs_id}/annotations/{annotation_type}",
    response_model_exclude_none=True,
    summary="Retrieve annotations for a VRS Object",
    description="Retrieve annotations for a VRS Object by VRS ID and annotation type",
)
def get_object_annotations(
    request: Request,
    vrs_id: Annotated[StrictStr, Path(..., description="VRS ID for VRS Object")],
    annotation_type: Annotated[StrictStr, Path(..., description="Annotation type")],
) -> GetAnnotationResponse:
    """Retrieve annotations for a VRS Object."""
    av: AnyVar = request.app.state.anyvar
    try:
        annotations = av.get_object_annotations(vrs_id, annotation_type)
    except ObjectNotFoundError as e:
        raise HTTPException(
            status_code=HTTPStatus.NOT_FOUND,
            detail=f"VRS Object {vrs_id} not found",
        ) from e
    return GetAnnotationResponse(annotations=annotations)


@objects_router.put(
    "/object/{vrs_id}/mappings",
    response_model_exclude_none=True,
    summary="Add mapping to a VRS Object",
    description="Provide a mapping to associate with a VRS object. The source and dest objects must be registered with AnyVar before adding mappings.",
)
def add_object_mapping(
    request: Request,
    vrs_id: Annotated[StrictStr, Path(..., description="VRS ID")],
    mapping_request: Annotated[
        AddMappingRequest, Body(description="Mapping to associate with the VRS Object")
    ],
) -> AddMappingResponse:
    """Store a mapping for a VRS Object"""
    av: AnyVar = request.app.state.anyvar
    source_vrs_obj: objects.VrsObject = _get_vrs_object(av, vrs_id)
    dest_vrs_id = mapping_request.dest_id
    dest_vrs_obj: objects.VrsObject = _get_vrs_object(av, dest_vrs_id)

    # Add the mapping to the database
    mapping: metadata.VariationMapping | None = None
    mapping_type = mapping_request.mapping_type
    try:
        mapping = metadata.VariationMapping(
            source_id=vrs_id, dest_id=dest_vrs_id, mapping_type=mapping_type
        )
        av.put_mapping(mapping)
    except ValueError as e:
        _logger.exception(
            "Failed to add mapping `%s` on variation `%s`",
            mapping_request,
            vrs_id,
        )
        raise HTTPException(
            status_code=HTTPStatus.UNPROCESSABLE_ENTITY,
            detail=f"Failed to add annotation: {mapping_request}. {e}",
        ) from e

    return AddMappingResponse(
        source_object=source_vrs_obj,
        source_object_id=vrs_id,
        dest_object=dest_vrs_obj,
        dest_object_id=dest_vrs_id,
        mapping_type=mapping_type,
    )


@objects_router.get(
    "/object/{vrs_id}/mappings/{mapping_type}",
    response_model_exclude_none=True,
    summary="Retrieve mappings for a VRS Object",
    description="Retrieve mappings for a VRS Object by ID and mapping type",
)
def get_object_mapping(
    request: Request,
    vrs_id: Annotated[StrictStr, Path(..., description="VRS ID for variation")],
    mapping_type: Annotated[
        metadata.VariationMappingType, Path(..., description="Mapping type")
    ],
) -> GetMappingResponse:
    """Retrieve mappings for a VRS Object."""
    av: AnyVar = request.app.state.anyvar
    try:
        mappings = av.get_object_mappings(vrs_id, mapping_type)
    except ObjectNotFoundError as e:
        raise HTTPException(
            HTTPStatus.NOT_FOUND,
            detail=f"VRS Object {vrs_id} not found",
        ) from e

    return GetMappingResponse(mappings=mappings)



"""Define routes for VCF annotation and ingestion."""

import asyncio
import datetime
import logging
import os
import pathlib
import tempfile
import uuid
from typing import Annotated

from fastapi import (
    APIRouter,
    BackgroundTasks,
    File,
    Path,
    Query,
    Request,
    Response,
    UploadFile,
    status,
)
from fastapi.responses import FileResponse

import anyvar
from anyvar.anyvar import AnyVar
from anyvar.restapi.schema import ErrorResponse, RunStatusResponse
from anyvar.translate.base import TranslatorConnectionError
from anyvar.vcf.ingest import (
    RequiredAnnotationsError,
    VcfRegistrar,
    register_existing_annotations,
)

try:
    import aiofiles  # noqa: I001
    from anyvar.queueing import celery_worker
    from billiard.exceptions import TimeLimitExceeded
    from celery.exceptions import WorkerLostError
    from celery.result import AsyncResult
except ImportError:
    aiofiles = None
    celery_worker = None
    TimeLimitExceeded = None
    WorkerLostError = None
    AsyncResult = None

_logger = logging.getLogger(__name__)

vcf_router = APIRouter()

# high side estimate for time is 500 variants per second
_expected_vrs_ids_per_second = int(
    os.getenv("ANYVAR_EXPECTED_VRS_IDS_PER_SECOND", "500")
)


async def _annotate_vcf_async(
    response: Response,
    vcf: UploadFile,
    for_ref: bool,
    allow_async_write: bool,
    assembly: str,
    add_vrs_attributes: bool,
    run_id: str | None,
) -> RunStatusResponse | ErrorResponse:
    """Annotate with VRS IDs asynchronously.  See `annotate_vcf()` for parameter definitions."""
    # if run_id is provided, validate it does not already exist
    if (
        not anyvar.anyvar.has_queueing_enabled()
        or not AsyncResult
        or not aiofiles
        or not celery_worker
    ):
        _logger.warning(
            "Async VCF annotation requested but not enabled (has_queueing_enabled=%s, AsyncResult=%s, aiofiles=%s, celery_worker=%s)",
            anyvar.anyvar.has_queueing_enabled(),
            AsyncResult,
            aiofiles,
            celery_worker,
            stack_info=True,
        )
        response.status_code = status.HTTP_400_BAD_REQUEST
        return ErrorResponse(
            error="Required modules and/or configurations for asynchronous VCF annotation are missing"
        )

    if run_id:
        existing_result = AsyncResult(id=run_id)
        if existing_result.status != "PENDING":
            response.status_code = status.HTTP_400_BAD_REQUEST
            return ErrorResponse(
                error=f"An existing run with id {run_id} is {existing_result.status}. Fetch the completed run result before submitting with the same run_id."
            )

    # write file to shared storage area with a directory for each day and a random file name
    async_work_dir = os.environ.get("ANYVAR_VCF_ASYNC_WORK_DIR", None)
    utc_now = datetime.datetime.now(tz=datetime.UTC)
    file_id = str(uuid.uuid4())
    input_file_path = pathlib.Path(
        f"{async_work_dir}/{utc_now.year}{utc_now.month}{utc_now.day}/{file_id}"
    )
    if not input_file_path.parent.exists():
        input_file_path.parent.mkdir(parents=True)
    _logger.debug(
        "writing working file for async run %s vcf to %s", run_id, input_file_path
    )

    vcf_site_count = 0
    newline_bytes = b"\n"
    async with aiofiles.open(input_file_path, mode="wb") as fd:
        while buffer := await vcf.read(1024 * 1024):
            vcf_site_count += buffer.count(newline_bytes)
            await fd.write(buffer)
    _logger.debug(
        "wrote working file for async run %s vcf to %s", run_id, input_file_path
    )
    _logger.debug("vcf site count of async run %s vcf is %s", run_id, vcf_site_count)

    # submit async job
    task_result = celery_worker.annotate_vcf.apply_async(
        kwargs={
            "input_file_path": str(input_file_path),
            "assembly": assembly,
            "for_ref": for_ref,
            "allow_async_write": allow_async_write,
            "add_vrs_attributes": add_vrs_attributes,
        },
        task_id=run_id,
    )
    _logger.info(
        "%s - async annotation run submitted for vcf with %s sites",
        task_result.id,
        vcf_site_count,
    )

    # set response headers
    response.status_code = status.HTTP_202_ACCEPTED
    response.headers["Location"] = f"/vcf/{task_result.id}"
    retry_after = max(
        1,
        round(
            (vcf_site_count * (2 if for_ref else 1)) / _expected_vrs_ids_per_second, 0
        ),
    )
    _logger.debug("%s - retry after is %s", task_result.id, str(retry_after))
    response.headers["Retry-After"] = str(int(retry_after))
    return RunStatusResponse(
        run_id=task_result.id,
        status="PENDING",
        status_message=f"Run submitted. Check status at /vcf/{task_result.id}",
    )


async def _annotate_vcf_sync(
    request: Request,
    response: Response,
    bg_tasks: BackgroundTasks,
    vcf: UploadFile,
    for_ref: bool,
    allow_async_write: bool,
    assembly: str,
    add_vrs_attributes: bool,
) -> FileResponse | ErrorResponse:
    """Annotate with VRS IDs synchronously.  See `annotate_vcf()` for parameter definitions."""
    av: AnyVar = request.app.state.anyvar
    registrar = VcfRegistrar(data_proxy=av.translator.dp, av=av)

    with tempfile.NamedTemporaryFile(delete=False, suffix=".vcf") as temp_in:
        contents = await vcf.read()
        temp_in.write(contents)
        temp_in_path = pathlib.Path(temp_in.name)

    with tempfile.NamedTemporaryFile(delete=False, suffix=".vcf") as temp_out:
        temp_out_path = pathlib.Path(temp_out.name)

    try:
        registrar.annotate(
            input_vcf_path=temp_in_path,
            output_vcf_path=temp_out_path,
            compute_for_ref=for_ref,
            assembly=assembly,
            vrs_attributes=add_vrs_attributes,
        )
    except (TranslatorConnectionError, OSError, ValueError):
        _logger.exception(
            "Encountered error during registration of VCF file %s", vcf.filename
        )
        response.status_code = status.HTTP_500_INTERNAL_SERVER_ERROR
        return ErrorResponse(error="VCF registration failed.")

    if not allow_async_write:
        _logger.info("Waiting for object store writes from API handler method")
        av.object_store.wait_for_writes()

    bg_tasks.add_task(_working_file_cleanup, temp_in_path)
    bg_tasks.add_task(_working_file_cleanup, temp_out_path)

    return FileResponse(temp_out_path)


@vcf_router.put(
    "/vcf",
    summary="Register alleles from a VCF",
    description="Provide a valid VCF. All reference and alternate alleles will be registered with AnyVar. The file is annotated with VRS IDs and returned.",
    response_model=None,
)
async def annotate_vcf(
    request: Request,
    response: Response,
    bg_tasks: BackgroundTasks,
    vcf: Annotated[UploadFile, File(..., description="VCF to register and annotate")],
    for_ref: Annotated[
        bool,
        Query(description="Whether to compute VRS IDs for REF alleles"),
    ] = True,
    allow_async_write: Annotated[
        bool,
        Query(
            description="Whether to allow asynchronous write of VRS objects to database",
        ),
    ] = False,
    assembly: Annotated[
        str,
        Query(
            pattern="^(GRCh38|GRCh37)$",
            description="The reference assembly for the VCF",
        ),
    ] = "GRCh38",
    add_vrs_attributes: Annotated[
        bool,
        Query(
            description="Whether to annotate with VRS attributes (start, stop, state, length, repeat subunit length) or just IDs"
        ),
    ] = False,
    run_async: Annotated[
        bool,
        Query(
            description="If true, immediately return a '202 Accepted' response and run asynchronously",
        ),
    ] = False,
    run_id: Annotated[
        str | None,
        Query(
            description="When running asynchronously, use the specified value as the run id instead generating a random uuid",
        ),
    ] = None,
) -> FileResponse | RunStatusResponse | ErrorResponse:
    """Register alleles from a VCF and return a file annotated with VRS IDs."""
    # If async requested but not enabled, return an error
    if run_async and not anyvar.anyvar.has_queueing_enabled():
        _logger.warning(
            "Async VCF annotation requested but not enabled (run_async=%s, has_queueing_enabled=%s)",
            run_async,
            anyvar.anyvar.has_queueing_enabled(),
            stack_info=True,
        )
        response.status_code = status.HTTP_400_BAD_REQUEST
        return ErrorResponse(
            error="Required modules and/or configurations for asynchronous VCF annotation are missing"
        )

    # ensure the temporary file is flushed to disk
    vcf.file.rollover()

    try:
        # Submit asynchronous run
        if run_async:
            return await _annotate_vcf_async(
                response=response,
                vcf=vcf,
                for_ref=for_ref,
                allow_async_write=allow_async_write,
                assembly=assembly,
                add_vrs_attributes=add_vrs_attributes,
                run_id=run_id,
            )
        # Run synchronously
        else:  # noqa: RET505
            return await _annotate_vcf_sync(
                request=request,
                response=response,
                bg_tasks=bg_tasks,
                vcf=vcf,
                for_ref=for_ref,
                allow_async_write=allow_async_write,
                assembly=assembly,
                add_vrs_attributes=add_vrs_attributes,
            )
    except Exception:
        _logger.exception("Unhandled error encountered error during VCF registration")
        raise


def _working_file_cleanup(file_path: str, missing_ok: bool = False) -> None:
    """Cleanup working files after successful completion of async task.

    :param input_file_path: path to VCF file
    """
    try:
        _logger.debug("removing working file %s", file_path)
        pathlib.Path(file_path).unlink(missing_ok=missing_ok)
    except Exception as e:  # noqa: BLE001
        _logger.warning("unable to remove working file %s: %s", file_path, str(e))


async def _ingest_annotated_vcf_sync(
    request: Request,
    bg_tasks: BackgroundTasks,
    vcf: UploadFile,
    assembly: str,
    allow_async_write: bool,
    require_validation: bool,
) -> FileResponse | ErrorResponse | None:
    """Ingest annotated VCF synchronously.  See `annotated_vcf()` for parameter definitions."""
    av: AnyVar = request.app.state.anyvar

    with tempfile.NamedTemporaryFile(delete=False, suffix=".vcf") as temp_in:
        contents = await vcf.read()
        temp_in.write(contents)
        temp_in_path = pathlib.Path(temp_in.name)

    conflicts_file = register_existing_annotations(
        av, temp_in_path, assembly, require_validation
    )

    if not allow_async_write:
        _logger.info("Waiting for object store writes from API handler method")
        av.object_store.wait_for_writes()

    bg_tasks.add_task(_working_file_cleanup, temp_in_path)
    if conflicts_file:
        bg_tasks.add_task(_working_file_cleanup, conflicts_file)
        return FileResponse(conflicts_file)
    return None


async def _ingest_annotated_vcf_async(
    response: Response,
    vcf: UploadFile,
    assembly: str,
    allow_async_write: bool,
    require_validation: bool,
    run_id: str | None,
) -> RunStatusResponse | ErrorResponse:
    """Ingest annotated VCF asynchronously.  See `annotated_vcf()` for parameter definitions."""
    if not anyvar.anyvar.has_queueing_enabled() or not aiofiles:
        _logger.warning(
            "Async VCF annotation requested but not enabled (has_queueing_enabled=%s, aiofiles=%s)",
            anyvar.anyvar.has_queueing_enabled(),
            aiofiles,
            stack_info=True,
        )
        response.status_code = status.HTTP_400_BAD_REQUEST
        return ErrorResponse(
            error="Required modules and/or configurations for asynchronous VCF annotation are missing"
        )

    # if run_id is provided, validate it does not already exist
    if run_id:
        existing_result = AsyncResult(id=run_id)
        existing_result_status = existing_result.status

        # explicitly delete to limit chances of deadlocks in the Redis client
        del existing_result

        if existing_result_status != "PENDING":
            response.status_code = status.HTTP_400_BAD_REQUEST
            return ErrorResponse(
                error=f"An existing run with id {run_id} is {existing_result_status}.  Fetch the completed run result before submitting with the same run_id."
            )

    async_work_dir = os.environ.get("ANYVAR_VCF_ASYNC_WORK_DIR", None)
    utc_now = datetime.datetime.now(tz=datetime.UTC)
    file_id = str(uuid.uuid4())
    input_file_path = pathlib.Path(
        f"{async_work_dir}/{utc_now.year}{utc_now.month}{utc_now.day}/{file_id}"
    )
    if not input_file_path.parent.exists():
        input_file_path.parent.mkdir(parents=True)
    _logger.debug("writing working file for async vcf to %s", input_file_path)

    vcf_site_count = 0
    async with aiofiles.open(input_file_path, mode="wb") as fd:
        while buffer := await vcf.read(1024 * 1024):
            vcf_site_count += buffer.count(b"\n")
            await fd.write(buffer)
    _logger.debug("wrote working file for async vcf to %s", input_file_path)
    _logger.debug("vcf site count of async vcf is %s", vcf_site_count)

    task_result = celery_worker.ingest_annotated_vcf.apply_async(
        kwargs={
            "input_file_path": str(input_file_path),
            "assembly": assembly,
            "allow_async_write": allow_async_write,
            "require_validation": require_validation,
        },
        task_id=run_id,
    )
    _logger.info(
        "%s - async annotation run submitted for vcf with %s sites",
        task_result.id,
        vcf_site_count,
    )

    # set response headers
    response.status_code = status.HTTP_202_ACCEPTED
    response.headers["Location"] = f"/vcf/{task_result.id}"
    # low side estimate for time is 333 variants per second
    retry_after = max(1, round((vcf_site_count * 2) / 333, 0))
    _logger.debug("%s - retry after is %s", task_result.id, str(retry_after))
    response.headers["Retry-After"] = str(retry_after)
    return RunStatusResponse(
        run_id=task_result.id,
        status="PENDING",
        status_message=f"Run submitted. Check status at /vcf/{task_result.id}",
    )


@vcf_router.put(
    "/annotated_vcf",
    summary="Register alleles from a VCF that has already been annotated with VRS objects",
    description="Provide a VCF that already has VRS location and state annotations. Ingest the objects into AnyVar.",
    response_model=None,
)
async def annotated_vcf(
    request: Request,
    response: Response,
    bg_tasks: BackgroundTasks,
    vcf: Annotated[
        UploadFile,
        File(
            ...,
            description="VCF that has already been annotated with VRS ID, start/stop, and state for all alleles",
        ),
    ],
    allow_async_write: Annotated[
        bool,
        Query(
            description="Whether to allow asynchronous write of VRS objects to database",
        ),
    ] = False,
    assembly: Annotated[
        str,
        Query(
            pattern="^(GRCh38|GRCh37)$",
            description="The reference assembly for the VCF",
        ),
    ] = "GRCh38",
    run_async: Annotated[
        bool,
        Query(
            description="If true, immediately return a '202 Accepted' response and run asynchronously",
        ),
    ] = False,
    require_validation: Annotated[
        bool,
        Query(
            description="If true, verify correctness of annotated ID and return CSV listing all validation failures"
        ),
    ] = False,
    run_id: Annotated[
        str | None,
        Query(
            description="When running asynchronously, use the specified value as the run id instead generating a random uuid",
        ),
    ] = None,
) -> FileResponse | RunStatusResponse | ErrorResponse | None:
    """Register alleles from a VCF and return a file annotated with VRS IDs."""
    # If async requested but not enabled, return an error
    if (run_async and not anyvar.anyvar.has_queueing_enabled()) or not AsyncResult:
        _logger.warning(
            "Async VCF annotation requested but not enabled (run_async=%s, has_queueing_enabled=%s)",
            run_async,
            anyvar.anyvar.has_queueing_enabled(),
            stack_info=True,
        )
        response.status_code = status.HTTP_400_BAD_REQUEST
        return ErrorResponse(
            error="Required modules and/or configurations for asynchronous VCF ingest are missing"
        )

    # ensure the temporary file is flushed to disk
    vcf.file.rollover()

    if run_async:
        if run_id:
            existing_result = AsyncResult(id=run_id)
            if existing_result.status != "PENDING":
                response.status_code = status.HTTP_400_BAD_REQUEST
                return ErrorResponse(
                    error=f"An existing run with id {run_id} is {existing_result.status}.  Fetch the completed run result before submitting with the same run_id."
                )
        return await _ingest_annotated_vcf_async(
            response=response,
            vcf=vcf,
            allow_async_write=allow_async_write,
            assembly=assembly,
            require_validation=require_validation,
            run_id=run_id,
        )
    try:
        return await _ingest_annotated_vcf_sync(
            request=request,
            bg_tasks=bg_tasks,
            vcf=vcf,
            assembly=assembly,
            allow_async_write=allow_async_write,
            require_validation=require_validation,
        )
    except RequiredAnnotationsError:
        _logger.exception("%s lacks required VRS annotations", vcf.filename)
        response.status_code = status.HTTP_400_BAD_REQUEST
        return ErrorResponse(
            error="Required VRS annotations are missing -- ensure INFO field has VRS_Allele_IDs, VRS_Starts, VRS_Ends, VRS_States, VRS_Lengths, and VRS_RepeatSubunitLengths"
        )
    except (TranslatorConnectionError, OSError, ValueError):
        _logger.exception(
            "Encountered error during registration of VCF file %s", vcf.filename
        )
        response.status_code = status.HTTP_500_INTERNAL_SERVER_ERROR
        return ErrorResponse(error="VCF ingestion failed.")


@vcf_router.get(
    "/vcf/{run_id}",
    summary="Poll for status and/or result for asynchronous VCF ingestion",
    description="Provide a valid run id to get the status and/or result of a VCF ingestion run",
    response_model=None,
)
async def get_vcf_run_status(
    response: Response,
    bg_tasks: BackgroundTasks,
    run_id: Annotated[
        str, Path(description="The run id to retrieve the result or status for")
    ],
) -> RunStatusResponse | FileResponse | ErrorResponse:
    """Return the status or result of an asynchronous registration of alleles from a VCF file."""
    # Asynchronous VCF annotation not enabled, return error
    if (
        not anyvar.anyvar.has_queueing_enabled()
        or not AsyncResult
        or not TimeLimitExceeded
        or not WorkerLostError
    ):
        _logger.warning(
            "Async VCF annotation requested but not enabled (has_queueing_enabled=%s, AsyncResult=%s, TimeLimitExceeded=%s, WorkerLostError=%s)",
            anyvar.anyvar.has_queueing_enabled(),
            AsyncResult,
            TimeLimitExceeded,
            WorkerLostError,
            stack_info=True,
        )
        response.status_code = status.HTTP_400_BAD_REQUEST
        return ErrorResponse(
            error="Required modules and/or configurations for asynchronous VCF annotation are missing"
        )

    # get the async result
    async_result = AsyncResult(id=run_id)
    _logger.debug("%s - status is %s", run_id, async_result.status)

    # completed successfully
    if async_result.status == "SUCCESS":
        response.status_code = status.HTTP_200_OK
        output_file_path = async_result.result
        async_result.forget()
        if output_file_path:
            _logger.debug("%s - output file path is %s", run_id, output_file_path)
            bg_tasks.add_task(_working_file_cleanup, output_file_path)
            return FileResponse(path=output_file_path)
        # for tasks that don't need to return a file, just send a success notification
        return RunStatusResponse(
            run_id=run_id,
            status="SUCCESS",
            status_message="VCF registration complete",
        )

    # failed - return an error response
    elif (  # noqa: RET505
        async_result.status == "FAILURE"
        and async_result.result
        and isinstance(async_result.result, Exception)
    ):
        # get error message and code
        error_msg = str(async_result.result)
        error_code = (
            "TIME_LIMIT_EXCEEDED"
            if isinstance(async_result.result, TimeLimitExceeded)
            else (
                "WORKER_LOST_ERROR"
                if isinstance(async_result.result, WorkerLostError)
                else "RUN_FAILURE"
            )
        )
        _logger.debug("%s - failed with error %s", run_id, error_msg)

        # cleanup working files
        if async_result.kwargs:
            input_file_path_str = async_result.kwargs.get("input_file_path", None)
            if input_file_path_str:
                input_file_path = pathlib.Path(input_file_path_str)
                if input_file_path.is_file():
                    _logger.debug(
                        "%s - adding task to remove input file %s",
                        run_id,
                        str(input_file_path),
                    )
                    bg_tasks.add_task(
                        _working_file_cleanup, str(input_file_path), missing_ok=True
                    )
                output_file_path = pathlib.Path(f"{input_file_path_str}_outputvcf")
                if output_file_path.is_file():
                    _logger.debug(
                        "%s - adding task to remove output file %s",
                        run_id,
                        str(output_file_path),
                    )
                    bg_tasks.add_task(
                        _working_file_cleanup, str(output_file_path), missing_ok=True
                    )

        # forget the run and return the response
        async_result.forget()
        response.status_code = int(
            os.environ.get("ANYVAR_VCF_ASYNC_FAILURE_STATUS_CODE", "500")
        )
        return ErrorResponse(error_code=error_code, error=error_msg)

    # status here is either "SENT" or "PENDING"
    else:
        # the after_task_publish handler sets the state to "SENT"
        #  so a status of PENDING is actually unknown task
        # but there can be a race condition, so if status is pending
        #  pause half a second at a time up to 5 seconds
        if async_result.status == "PENDING":
            for _ in range(10):
                await asyncio.sleep(0.5)
                _logger.debug(
                    "%s - after 0.5 second wait, status is %s",
                    run_id,
                    async_result.status,
                )
                if async_result.status != "PENDING":
                    break

        # status is "PENDING" - unknown run id
        if async_result.status == "PENDING":
            response.status_code = status.HTTP_404_NOT_FOUND
            return RunStatusResponse(
                run_id=run_id,
                status="NOT_FOUND",
                status_message="Run not found",
            )
        # status is "SENT" - return 202
        #  with retry after 2 seconds
        else:  # noqa: RET505
            response.status_code = status.HTTP_202_ACCEPTED
            response.headers["Retry-After"] = "2"
            return RunStatusResponse(
                run_id=run_id,
                status="PENDING",
                status_message=f"Run not completed. Check status at /vcf/{run_id}",
            )
